# x-glcs.sml
#  _______________________________________________________________________
# |									|
# |			      Projet LCS				|
# |			  Groupe OLC, CNRS/LAAS				|
# |			 7, Av. du Colonel Roche			|
# |			   31077 Toulouse Cedex				|
# |_______________________________________________________________________|
# 
# graphic interface for glcs simulator, tcl/tk part
#
# Version 5.0; June 98;
# 
#




#######################################################################
# test (local) or slcs mode
#######################################################################

set debug false
# set debug true



#######################################################################
# interface entre slcs et arbre.tcl
#######################################################################

# --- commands from slcs to interface:

# variable slcs may be used (though vwait) to wait command completion
set slcs true

# create root node:
proc root {} {
    global current slcs
    New
    hidetag $current
    set slcs true
}

# add son to current node:
proc new {t} {
    global current slcs aa nn
    mktree $current
    global v$current.t
    set v$current.t $t
    if {$aa == "shown"} {showtag $current.t} else {hidetag $current.t}
    if {$nn == "shown"} {showtag $current} else {hidetag $current}
    set slcs true
}

# move current node, relative:
proc left {n} {global slcs; Left $n; set slcs true}
proc right {n} {global slcs; Right $n; set slcs true}
proc lmost {} {global slcs; Lmost; set slcs true}
proc rmost {} {global slcs; Rmost; set slcs true}

# tags current node with t:
proc tag {t} {
    global current slcs
    global v$current
    set v$current $t
    showtag $current
    set slcs true
}

# nop:
proc more {} {}
proc done {} {update idletasks; showcurrent}

# set properties of current node:
proc dead {} {paintCurrent grey75}
proc fail {} {paintCurrent black}

# set parameters for interface -> slcs dialog:
proc steps {n} {global ss; set ss $n}
proc mode {m} {global mm; set mm $m}


# --- commands from interface to slcs:
# commands sent from buttons are echoed by slcs

proc write {s} {puts $s; flush stdout}

# uinterrupt (bound to ^C in text)
proc uinterrupt {} {
    write "I"
}

# uquit (bound to quit button): 
proc uquit {} {
    global debug
    if {$debug != "true"} {
	write "Q"
    }
    exit
}

# utag (send tag to slcs):
proc utag {p t} {
    global current slcs
    set back $current
    if {$p != $back} {set slcs false; umove $p; vwait slcs}
    set slcs false; write "Ctag \"$t\"\;"; vwait slcs
    if {$p != $back} {set slcs false; umove $back; vwait slcs}
}

# umove (bound to <1>)
proc umove {p} {
    global debug current roott
    if {$debug == "true"} {
	mkcurrent $p
	showcurrent
    } else {
	set old $current
	# compute displacement n from old to p and issue move command:
	set n 0
	set lm [golmost [lindex [sons $roott] 0]]
	while {$lm != $old && $lm != $p} {
	    set lm [goright $lm]
	}
	if {$lm == $old} {
	    set side "Cright "
	    while {$lm != $p} {incr n; set lm [goright $lm]}
	} else {	
	    set side "Cleft "
	    while {$lm != $old} {incr n; set lm [goright $lm]}
	}
	if {$n != 0} {
	write $side$n\;
	}
    }
}

# umore (bound to double <1>)
proc umore {p} {
    global debug
    if {$debug == "true"} {
	mktree $p
    } else {
	write "Cmore();"
    }
}

# uextractmore (bound to double control <1>)
proc uextractmore {p} {write "Ctry next;"}

# uright, uleft, ulmost, urmost (bound to arrow keys)
proc uleft {} {write "Cleft 1;"}
proc uright {} {write "Cright 1;"}
proc ulmost {} {write "Clmost ();"}
proc urmost {} {write "Crmost ();"}

# usteps (bound to steps button)
proc usteps {n} {
    global ss ssbak
    if {"" == [string trimleft $n 0123456789]} {
	if {$ss != $ssbak} {write "Csteps $n;"}
    } else {
	bell; set ss $ssbak
    }
}

# umode (bound to mode button)
proc umode {m} {
    global mm
    if {$m != $mm} {write "Cmode \"$m\";"}
}



#######################################################################
# tree editor (based on TreeEdit, without remove and show)
#######################################################################

# wm title . "glcs behavior stepper"
wm title . glcs
wm geometry . 650x750

set padd 5

# --- top frame, canvas and drawing frame:
# note: width follows from geometry
proc ScrolledCanvas {c width height region} {
    frame $c -bg white
    canvas $c.c -width $width -height $height \
	        -scrollregion $region \
		-xscrollcommand [list $c.xscroll set] \
		-yscrollcommand [list $c.pad.yscroll set] \
		-borderwidth 3 -bg white -highlightthickness 0 
    frame $c.pad
    scrollbar $c.pad.yscroll -orient vertical -command [list $c.c yview]
    set padw [expr [$c.pad.yscroll cget -width] + 2* \
	    ([$c.pad.yscroll cget -borderwidth] + \
	     [$c.pad.yscroll cget -highlightthickness])]
    frame $c.pad.pad -height $padw -width $padw
    pack $c.pad.pad -side bottom
    pack $c.pad.yscroll -side bottom -fill y -expand true
    pack $c.pad -side right -fill y
    scrollbar $c.xscroll -orient horizontal -command [list $c.c xview]
    pack $c.xscroll -side bottom -fill x
    pack $c.c -side left -fill both -expand true -padx 5 -pady 5
    pack $c -side top -fill both -expand true
    $c.c xview moveto 0.0
    $c.c yview moveto 0.0
    return $c.c
}
frame .t -borderwidth 3 -bg white -relief sunken 
pack .t -side top -fill both -padx $padd -pady $padd 
# -expand true  [FIXED SIZE]
set roott [ScrolledCanvas .t.u  600 450 {0 0 1000 1000}]
pack $roott -side top -fill both -expand true -padx 5 -pady 5
$roott create window 0 0 -anchor nw -window [frame $roott.s -bg white]
# the tree is drawn in $roott.s (sons of roott)
bind $roott <1> {showcurrent}
bind $roott <Double-1> {home}
bind all <Left> {uleft}
bind all <Right> {uright}
bind all <Down> {ulmost}
bind all <Up> {urmost}



# --- buttons
frame .b -cursor dotbox
frame .b.cmd -cursor top_left_arrow -borderwidth 0 -highlightthickness 0

frame .b.cmd.steps
set ss 5
button .b.cmd.steps.ss -width 4 -text Steps -command {focus -force .b.cmd.steps.ee}
entry .b.cmd.steps.ee -bg white -textvariable ss -justify right -relief flat -width 4
pack .b.cmd.steps.ss .b.cmd.steps.ee -side left
bind .b.cmd.steps.ee <FocusIn> {set ssbak $ss}
bind .b.cmd.steps.ee <Return> {focus -force $buffer; usteps $ss}

frame .b.cmd.mode
set mm "xpand"
menubutton .b.cmd.mode.mm -width 6 -text Mode -menu .b.cmd.mode.mm.menu -relief raised
label .b.cmd.mode.ee -bg white -textvariable mm -width 5
pack .b.cmd.mode.mm .b.cmd.mode.ee -side left
menu .b.cmd.mode.mm.menu -tearoff 0
.b.cmd.mode.mm.menu add command -label run -command {umode "run"}
.b.cmd.mode.mm.menu add command -label xpand -command {umode "xpand"}
.b.cmd.mode.mm.menu add command -label scan -command {umode "scan"}


frame .b.cmd.atags
set aa "shown"
proc chatag {} {
    global aa
    if {$aa == "shown"} {
	set aa "hidden"; map hideatag
    } else {
	set aa "shown"; map showatag
    }
    showcurrent
}
button .b.cmd.atags.aa -width 4 -text Actions -command {chatag}
label .b.cmd.atags.ee -bg white -textvariable aa -width 6
pack .b.cmd.atags.aa .b.cmd.atags.ee -side left

frame .b.cmd.ntags
set nn "hidden"
proc chntag {} {
    global nn
    if {$nn == "shown"} {
	set nn "hidden"; map hidentag
    } else {
	set nn "shown"; map showntag
    }
    showcurrent
}
button .b.cmd.ntags.nn -width 4 -text Tags -command {chntag}
label .b.cmd.ntags.ee -bg white -textvariable nn -width 6
pack .b.cmd.ntags.nn .b.cmd.ntags.ee -side left

# help window
proc help {} {
    # raises help window if already open
    if {[winfo exist .h]} {
	raise .h .
    } else {
	toplevel .h
	wm title .h help
	message .h.msg -bg white \
		-font lucidasans-12 \
		-width 420 \
		-text "\n \
		\Mouse controls on nodes:\n \
		\    Left makes node the resumption\n \
		\    Double-Left expands node\n \
		\    Triple-Left makes node root and expands\n \
		\    Middle folds/unfolds tree\n \
		\    Right shows or hides node tag\n \
		\    Control-Right shows or hides action\n\n \
	        \Tag updates:\n \
		\    Validate by <Return> or <Tab>\n\n \
	        \Arrows controls:\n \
		\    Left moves resumption one position left\n \
		\    Right moves resumption one position right\n \
		\    Down moves resumption to leftmost position\n \
		\    Up moves resumption to rightmost position\n\n \
		\Mouse controls on backgroud:\n \
		\    Left scrolls so that resumption visible\n \
		\    Double-left scrolls to home position\n"
	button .h.ok -text Dismiss -command {destroy .h}
	pack .h.msg -side top -fill both -expand true
	pack .h.ok -side top -fill x
    }
}
button .b.cmd.help -text Help -width 4 -command {help}

pack .b.cmd.steps -side left
pack .b.cmd.mode .b.cmd.atags .b.cmd.ntags .b.cmd.help -padx 5 -side left
pack .b.cmd -side left -padx $padd

frame .b.bye -cursor top_left_arrow -borderwidth 1 -relief sunken -highlightthickness 0
button .b.bye.quit -text Quit -width 6 -command {uquit}
pack .b.bye.quit -side right
pack .b.bye -side right -padx $padd
pack .b -side top -fill x


# procedure and bindings to resize drawing canvas by dragging buttons
set yfrom 0
proc dragButtons {yto} {
    global yfrom roott text
    if {$yfrom!=0} {
	set dy [expr $yto - $yfrom]
	if {[expr [winfo height $text] - $dy] > 40} {
	    set yfrom $yto
	    $roott configure -height [expr [$roott cget -height] + $dy]
	    $text see end
	}
    }
}
bind .b <1> {set yfrom %Y}
bind .b <B1-Motion> {dragButtons %Y}



# --- text window

# text area, adjustable size, non editable:
frame .e -borderwidth 3 -bg white -relief sunken
pack .e -side top -fill both -padx $padd -pady $padd -expand true
set text [text .e.t -state disabled -borderwidth 0 -bg white -highlightthickness 0 -yscrollcommand ".e.b set"]
scrollbar .e.b -orient vertical -command [list .e.t yview]
pack .e.b -side right -fill y
pack .e.t -side left -fill both -padx 10 -expand true 

# one-line text buffer, embedded into text window, before end character:
set buffer [entry .e.u -textvariable cont -width 0 -bg white -borderwidth 0 -highlightthickness 0]
if {$debug == "true"} {$buffer configure -bg lightgrey} else {}
$text window create end -window $buffer
focus $buffer

# paste always in buffer, line by line:
proc paste {} {
    global text buffer
    # catch prevents tcl error when paste with no selection
    if {0 == [catch {set line [selection get -displayof $text]}]} {
	while {[set i [string first \n $line]] != -1} {
	    set fst [string range $line 0 [expr $i -1]]
	    set line [string range $line [expr $i + 1] [expr [string length $line] -1]]
	    $buffer insert end $fst
	    emit
	}
	$buffer insert end $line
    }
}
bind $text <ButtonRelease-2> {paste}

# bindings to send text buffer to slcs
bind $buffer <Return> {emit}
bind $buffer <Control-Key-d> {uquit}
bind $buffer <Control-Key-c> {uinterrupt}

# sends a comment line to slcs (useful for debugging interface ..)
proc comment {l} {
    global cont
    set cont "(* $l *);"
    emit
}
# send line to slcs:
proc emit {} {
    global text cont buffer
    # send buff to slcs
    write L$cont
    # copy buffer in text
    $text configure -state normal
    $text insert "end -2 chars" $cont
    $text insert "end -2 chars" "\n"
    $text configure -state disabled
    set cont ""
    $text see end
    focus $buffer
}

# proc to append data received from lcs into text:
proc e {l} {
    global text
    $text configure -state normal
    $text insert "end -2 chars" $l
    $text configure -state disabled
    $text see end
}


# --- building trees

# colours of arcs
set visible red
set hidden white

# $x ends with .number, returns number
proc last {x} {
    return  [string range $x \
            [expr 1 + [string length [string trimright $x .0123456789]]] \
	    [string length $x]]
}

proc father {p} {
    return [string trimright [string trimright $p .0123456789] .s]
}

proc sons {p} {
    return [pack slaves $p.s]
}

# rest value for current and fcurrent
set current $roott
set fcurrent $roott

# makes node $p current ($current assumed materialized)
proc mkcurrent {p} {
    global fcurrent current roott
    # unmark $current and fcurrent
    if {$current != $roott} {
	$current.n raise disk
    }
    if {$fcurrent != $roott} {
	$fcurrent.fn raise disk
    }
    set current $p
    set fcurrent $roott
    # showcurrent will mark current and fcurrent
}

proc paintCurrent {c} {
    global current
    $current.n itemconfigure disk -fill $c -outline $c
}


# performs f n on all nodes: usage: map f
proc rmap {f w} {
    set l [pack slaves $w]
    set n [llength $l]
    while {$n > 0} {
	set w [lindex $l [expr $n - 1]]
	eval "$f $w"
	rmap $f $w.s
	set n [expr $n - 1]
    }
}
proc map {f} {global roott; rmap $f $roott.s}


# --- updates scroll so that current node is visible
proc delta {x min max} {
    if {$x < $min} {return [expr $x - $min]}
    if {$x > $max} {return [expr $x - $max]}
    return 0
}
proc position {p} {
    # returns coordinates of center of $p.n, relative to origin of view area
    global roott fcurrent
    set x [expr 10 + [winfo x $p.n]]
    set y [expr 10 + [winfo y $p.n]]
    while {$p != $roott} {
	# if p folded, replace current by folded:
	if {-1 == [lsearch [pack slaves $p] $p.n]} {
	    set x [expr 10 + [winfo x $p.fn]]
	    set y [expr 10 + [winfo y $p.fn]]
	    if {$fcurrent != $roott} {$fcurrent.fn raise disk}
	    set fcurrent $p
	    $p.fn raise ring
	}
	set q [string trimright $p .0123456789]
	set x [expr $x + [winfo x $p] + [winfo x $q]]
	set y [expr $y + [winfo y $p] + [winfo y $q]]
	set p [string trimright $q .s]
    }
    set x [expr $x + [winfo x $p]]
    set y [expr $y + [winfo y $p]]
    return "$x $y"
}
proc showcurrent {} {
    # updates scroll so that current is visible
    global current roott
    # necessary:
    update idletasks
    if {$current == $roott} {return}
    # mark current
    $current.n raise ring
    # get position of node, relative to view origin:
    set pos [position $current]
    set x [lindex $pos 0]
    set y [lindex $pos 1]
    # compute deltas:
    set dx [delta $x 32 [expr [winfo width $roott] - 32]]
    set dy [delta $y 32 [expr [winfo height $roott] - 32]]
    if {$dx == 0 && $dy == 0} return
    # get scroll span:
    set r [$roott cget -scrollregion]
    set xs [lindex $r 2]
    set ys [lindex $r 3]
    # get origin of view:
    # set x0 [expr $xs * [lindex [$roott xview] 0]]
    # set y0 [expr $ys * [lindex [$roott yview] 0]]
    set x0 [expr 3 - [winfo x $roott.s]]
    set y0 [expr 3 - [winfo y $roott.s]]
    # enlarge scrollregion if $current out of region:
    set rr $r
    if {[expr $x + $x0 + 50] > $xs} {
	set rr [lreplace $rr 2 2 [set xs [expr $x + $x0 + 1000]]]
    }
    if {[expr $y + $y0 + 50] > $ys} {
	set rr [lreplace $rr 3 3 [set ys [expr $y + $y0 + 1000]]]
    }
    if {$rr != $r} {$roott configure -scrollregion $rr}
    # adjust scroll:
    if {$dx != 0} {$roott xview moveto [expr [expr $x0 + $dx.0] / $xs]}
    if {$dy != 0} {$roott yview moveto [expr [expr $y0 + $dy.0] / $ys]}
}
proc home {} {
    global roott
    $roott xview moveto 0.0
    $roott yview moveto 0.0
}

# creates root of tree
proc New {} {
    global current roott
    set tree [pack slaves $roott.s]
    if {0 != [llength $tree]} {set current $roott; destroy $tree}
    mktree $roott
}

# adds son to $p; adds a tree in frame $p.s
proc mktree {p} {
    global roott visible hidden

    # configuring to-bar frames
    set lbg $hidden
    set rbg $hidden
    set tbg $visible
    set pad 0
    if {$p != $roott} {
	set sons [sons $p]
	set l [llength $sons]
	switch $l {
	    0 {
		# make $p's from bar visible
		$p.f configure -bg $visible
	    }
	    1 {
		# make right h-bar of leftmost son visible, hide its 
		# left h-bar and make current left h-bar visible
		set first [lindex $sons 0]
		$first.h.l configure -bg $hidden
		$first.h.r configure -bg $visible
		set lbg $visible
	    }
	    default {
		# make right h-bar of previous, and current left h-bar, visible
		[lindex $sons [expr $l - 1]].h.r configure -bg $visible
		set lbg $visible
	    }
	}
    } else {
	set pad 10
	set tbg $hidden
    }

    # computing an index for tree (1 + index of youngest brother if any, or 0)
    set me 0
    set sibs [sons $p]
    set len [llength $sibs]
    if {$len > 0} {set me [expr 1 + [last [lindex $sibs [expr $len - 1]]]]}

    # subtree frame
    set w $p.s.$me
    frame $w -bg white -bd 0 -highlightthickness 0
    pack $w -side left -padx $pad -fill y
    # node and bars
    # h-bar
    frame $w.h -bd 0 -highlightthickness 0
    frame $w.h.l -width 0 -height 4 -bg $lbg -bd 0 -highlightthickness 0
    frame $w.h.m -width 4 -height 4 -bg $tbg -bd 0 -highlightthickness 0
    frame $w.h.r -width 0 -height 4 -bg $rbg -bd 0 -highlightthickness 0
    pack $w.h.l -side left -expand true -fill x
    pack $w.h.m -side left
    pack $w.h.r -side left -expand true -fill x
    pack $w.h -side top -fill x
    # to bar (in three pieces for possible label insertion in $w.t.l)
    frame $w.t -bg white -bd 0 -highlightthickness 0
    frame $w.t.n -width 4 -height 5 -bg $tbg -bd 0 -highlightthickness 0
    frame $w.t.b -width 4 -height 5 -bg $tbg -bd 0 -highlightthickness 0
    mktag $w.t red #ffeeee ""
    pack $w.t.n $w.t.b -side top
    pack $w.t -side top
    # node
    canvas $w.n -width 21 -height 21 -bg white -bd 0 -highlightthickness 0
    $w.n create oval 0 0 20 20 -fill blue -outline blue -tag disk
    $w.n create oval 2 2 18 18 -width 2 -outline white -tag ring
    mktag $w blue #eeeeff "" 
    pack $w.n -side top -padx 4
    # from bar
    frame $w.f -width 4 -height 5 -bg $hidden -bd 0 -highlightthickness 0
    pack $w.f -side top
    # sons frame
    frame $w.s -bg white -bd 0 -highlightthickness 0
    pack $w.s -side top 

    # bindings

    $w.n bind all <1> {umove [string trimright %W .n]}
    $w.n bind all <Double-1> {umore [string trimright %W .n]}
    $w.n bind all <Triple-1> {uextractmore [string trimright %W .n]}
    $w.n bind all <2> {fold [string trimright %W .n]}
    $w.n bind all <3> {switchtag [string trimright %W .n]}
    $w.n bind all <Control-3> {switchtag [string trimright %W .n].t}

    # makes current the last node created
    mkcurrent $w

    return $w 
}



# --- node folding routines

proc unfold {p} {
    # unfolding
    global fcurrent roott nn
    destroy $p.fn
    # update fcurrent, otherwise pb with showcurrent
    if {$fcurrent == $p} {set fcurrent $roott}
    pack $p.n -side top -after $p.t
    pack $p.f $p.s -side top
    if {"shown" == $nn} {showtag $p}
    showcurrent
}

proc fold {p} {
    # folding
    pack forget $p.n $p.s $p.f
    if {-1 != [lsearch [pack slaves $p] $p.l]} {pack forget $p.l}
    canvas $p.fn -width 21 -height 21 -bg white -highlightthickness 0
    pack $p.fn -side top -padx 4 -after $p.t
    $p.fn create poly 10 5 16 15 4 15 -width 2 -outline white -fill blue -tag ring
    $p.fn create poly 10 0 20 18 0 18 -fill blue -tag disk
    $p.fn bind all <1> {bell}
    $p.fn bind all <3> {bell}
    $p.fn bind all <2> "unfold $p"
    showcurrent
}



# --- nodes/arcs tagging routines

# if tag shown then hides it else shows it
proc switchtag {p} {
    if {-1 == [lsearch [pack slaves $p] $p.l]} {
	# hidden
	showtag $p
    } else {
	# shown
	hidetag $p
    }
}

# creates tag of $p with text $t
proc mktag {p bg fg t} {
    # to save space, creation of entry could be delayed until updated ...
    global v$p text buffer
    frame $p.l -bg $bg -highlightthickness 0 -borderwidth 0
    # pack $p.l -side top -after $p.n -padx 1
    set v$p $t
    entry $p.l.e -textvariable v$p -bg $fg -font lucidasans-10 \
	    -justify center -relief flat -width 0 -highlightthickness 0
    pack $p.l.e -padx 1 -pady 1
    bind $p.l.e <FocusIn> "savetag $p"
    bind $p.l.e <Return> "focus -force $buffer; updtag $p"
    bind $p.l.e <Tab> "updtag $p"
}

# saves current tag into oldtag variable
proc savetag {p} {
    global oldtag v$p
    grab $p.l
    set oldtag [set v$p]
}

# invoked on tag changes to update slcs tags
proc updtag {p} {
    global v$p oldtag
    set newtag [set v$p]
    grab release $p.l
    # send new tag to slcs, if node tag and if changed:
    if {[string trimright $p .t] == $p} {
	if {$oldtag != $newtag} {
	    utag [string trimright $p .n] $newtag
	}
    }
}

# hides tag of $p
proc hidetag {p} {
    pack forget $p.l
}
proc hidentag {p} {if {-1 != [lsearch -exact [pack slaves $p] $p.n]} {hidetag $p}}
proc hideatag {p} {hidetag $p.t}

# shows tag of $p
proc showtag {p} {
    pack $p.l -side top -after $p.n -padx 1
}
proc showntag {p} {if {-1 != [lsearch -exact [pack slaves $p] $p.n]} {showtag $p}}
proc showatag {p} {global roott; set r $roott.s.0; if {$p != $r} {showtag $p.t}}


# --- preorder traversal routines

# makes lowest current (in preorder traversal)
proc Lmost {} {
    global roott
    mkcurrent [golmost [lindex [sons $roott] 0]]
    showcurrent
}

# makes highest current (in preorder traversal)
proc Rmost {} {
    global roott
    set l [sons $roott]
    mkcurrent [lindex $l [expr [llength $l] - 1]]
    showcurrent
}

# makes successor current (in preorder traversal)
proc Right {n} {
    global current
    set result $current
    while {$n > 0} {incr n -1; set result [goright $result]}
    mkcurrent $result
    showcurrent
}

# makes predecessor current (in preorder traversal)
proc Left {n} {
    global current
    set result $current
    while {$n > 0} {incr n -1; set result [goleft $result]}
    mkcurrent $result
    showcurrent
}

# returns predecessor of $c (in preorder traversal)
proc goleft {c} {
    global roott
    set s [sons $c]
    set l [llength $s]
    if {$l > 0} {
	# my right son
	lindex $s [expr $l - 1]
    } else {
	# the left sibling of the youngest ancestor of $c that has one, or $c
	set f [father $c]
	set sibs [sons $f]
	set me [lsearch $sibs $c]
	while {$me == 0} {
	    if {$f == $roott} {
		return $c
	    } else {
		set from $f
		set f [father $f]
		set sibs [sons $f]
		set me [lsearch $sibs $from]
	    }
	}
	lindex $sibs [expr $me - 1]
    }
}

# returns successor of $c (in preorder traversal)
proc goright {c} {
    global roott
    # lmost of right sibling, orelse if I'am roott then me else my father
    set f [father $c]
    set sibs [sons $f]
    set next [expr 1 + [lsearch $sibs $c]]
    if {$next >= [llength $sibs]} {
	if {$f == $roott} {
	    return $c
	} else {
	    return $f
	}
    } else {
	golmost [lindex $sibs $next]
    }   
}

# returns smallest in subtree $c
proc golmost {c} {
    set s [sons $c]
    if {0 == [llength $s]} {return $c} else {golmost [lindex $s 0]}
}




# --- miscellaneous




# --- initial state

if {$debug == "true"} {New}

