(* fun.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

*)






(* interface

structure Fun =

   abstype 'a -> 'b
   val I = - : 'a -> 'a
   val o = - : ('a -> 'b) * ('c -> 'a) -> 'c -> 'b
   val fun_pow = - : int -> ('a -> 'a) -> 'a -> 'a
   val repeat = - : int -> ('a -> 'b) -> 'a -> unit

   val curry = - : ('a * 'b -> 'c) -> 'a -> 'b -> 'c
   val pair = - : ('a -> 'b -> 'c) -> 'a * 'b -> 'c
   val can = - : ('a -> 'b) -> 'a -> bool
   exception Assert : exn
   val assert = - : ('a -> bool) -> 'a -> 'a

   val fold = - : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
   val rev_fold = - : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
   exception Reduce : exn
   val reduce = - : ('a * 'a -> 'a) -> 'a list -> 'a
   val rscan = - : ('a * 'a -> 'a) -> 'a list -> 'a list

*)



(* implementation *)

structure Fun =
struct

open Core.Fun;

(* some useful combinators *)

infix 3 o;
fun (f o g) x = f (g x);
fun fun_pow n f x = if n=0 then x else fun_pow (n-1) f (f x);
fun repeat n f x = if n=0 then () else (f x; repeat (n-1) f x);
fun I x = x;
fun curry f x y = f (x,y);
fun pair f (x,y) = f x y;
fun can f x = (f x; true) handle Interrupt => raise Interrupt | _=> false;
exception Assert;
fun assert p x = if p x then x else raise Assert;

(* some list iterated function compositions *)
 
fun fold f = (* fold f [x1, ... ,xn] b = f x1 (f x2 ( ... (f xn b) ...) *)
        let fun fld [] = (fn b => b)
              | fld (h::t) = (fn b => f h (fld t b))
        in fld
        end;
fun rev_fold f = (* rev_fold f b [x1, ... ,xn] = (f ... (f (f b x1) x2) ...) xn) *)
        let fun fld u [xn] = f u xn
              | fld u (h::t) = fld (f u h) t
              | fld u [] = u
        in fld
        end;
exception Reduce;
fun reduce f = (* APL reduce: reduce f [x1,..,xn] = f (x1, f (x2,  ... xn ...)) *)
        let fun red [x] = x
              | red (h::t) = f (h,red t)
              | red [] = raise Reduce
        in red
        end;
fun rscan f = (* APL scan: scan f [x1,x2,...] = x1:: reduce f [x1,x2] :: ... *)
        let fun scn _ [] = []
              | scn [u] (h::t) = let val x = f (u,h) in x :: scn [x] t end
              | scn [] (h::t) = h :: scn [h] t
        in scn []
        end;
end;











