(* list.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

lists

*)




(* interface

type 'a list
con nil : 'a list
con :: : 'a * 'a list -> 'a list
exception Hd : exn
val hd : 'a list -> 'a
exception Tl : exn
val tl : `a list -> 'a list
val null : `a list -> bool
val length : 'a list -> int
val @ : 'a list * 'a list -> 'a list
val mem : 'a -> 'a list -> bool
val rev : 'a list -> 'a list
val flat : 'a list list -> 'a list
val map : ('a -> 'b) -> 'a list -> 'b list
val filter : ('a -> bool) -> 'a list -> 'a list
exception Find : exn
val find : ('a -> bool) -> 'a list -> 'a
val split : ('a * 'b) list -> 'a list * 'b list
exception Combine : exn
val combine : 'a list * 'b list -> ('a * 'b) list
val string_of_list : string * string * string -> ('a -> string) -> 'a list -> string

*)



(* equivalent definition

infixr 5 ::;
datatype 'a list = nil | :: of 'a * 'a list;
exception Hd;
fun hd [] = raise Hd | hd (h::_) = h;
exception Tl;
fun tl [] = raise Tl | tl (_::t) = t; 
fun null [] = true | null _ = false; 
*)



(* implementation *)

structure List =
struct

infixr 5 ::;
infixr 5 @;

open Core.List;

fun length l = 
	let fun len (k,nil) = k | len (k, a::x) = len (k+1,x) 
	in len (0,l) 
	end;
fun length [] = 0 | length (_::t) = 1 + length t;

(* general purpose *)
fun [] @ y = y | (h::t) @ y = h::(t @ y);
fun append x y = x @ y;
fun mem x =
  	let fun rmem [] = false | rmem (h::t) = (x=h) orelse rmem t
  	in rmem
  	end;
fun mem x [] = false | mem x (h::t) = (h=x) orelse mem x t;

local
fun rrev ([],b) = b | rrev (h::t,b) = rrev (t,h::b)
in
fun rev l = rrev (l,[])
end;
fun flat [] = [] | flat (h::t) = h @ flat t;
fun map f = let fun rmap [] = [] | rmap (h::t) = f h :: rmap t in rmap end;
fun map f [] = [] | map f (h::t) = f h:: map f t;
fun filter p [] = [] | filter p (h::t) = if p h then h:: filter p t else filter p t;
exception Find;
fun find p [] = raise Find | find p (h::t) = if p h then h else find p t;

(* pair of lists <-> list of pairs *)
fun split [] = ([],[])
  | split ((x,y)::t) = (let val (l1,l2) = split t in ((x::l1), (y::l2)) end);
exception Combine;
fun combine ([],[]) = []
  | combine (h1::t1, h2::t2) = (h1,h2):: combine (t1,t2)
  | combine _ = raise Combine;

(* for building list printers or unparsers *)
fun string_of_list (begs,seps,ends) u l =
	let fun sol [] = [ends]
	      | sol [h] = u h:: [ends]
	      | sol (h::t) = u h:: seps ::sol t
	in implode (begs :: sol l)
	end;
	
end;


