(* lizt.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

lazy lists.

*)




(* signature lizt (lazy lists)

datatype lizt
con nils : 'a lizt
con ::: : 'a * 'a lizt lift -> 'a lizt
exception Hs
val hs : 'a lizt -> 'a
exception Ts
val ts : 'a lizt -> 'a lizt
val nulls : 'a lizt -> bool
val maps : ('a -> 'b) -> 'a lizt -> 'b lizt
val filters : ('a -> bool) -> 'a lizt -> 'a lizt
val nths : 'a lizt -> int -> 'a
val splits : ('a * 'b) lizt -> 'a lizt * 'b lizt
exception Combines
val combines : 'a lizt * 'b lizt -> ('a * 'b) lizt
val generates : ('a -> 'a) -> 'a -> 'a lizt
*)



(* implementation *)



structure Lizt =
struct

infixr 5 :::;

datatype 'a lizt = nils | ::: of 'a * ('a lizt) lift;

exception Hs and Ts;
fun hs nils = raise Hs | hs (x ::: _) = x
and ts nils = raise Ts | ts (_ ::: up l) = l 
and nulls nils = true | nulls _ = false;

fun maps f nils = nils
  | maps f (x ::: up l) = f x ::: up (maps f l);

fun filters p nils = nils
  | filters p (x ::: up l) = if p x then x ::: up (filters p l) else filters p l;

exception Nths;
local fun nthrec (h ::: up t) = fn 1 => h | n => nthrec t (n-1)
in fun nths nils _ = raise Nths
     | nths l n = if n<0 then raise Nths else nthrec l n
end;

fun splits nils = (nils, nils)
  | splits ((x,y) ::: up t) = (* recursive call of splits must be delayed *)
	(let val up (l1,l2) = up (splits t) in ((x ::: up l1), (y ::: up l2)) end);

exception Combines
fun combines (nils,nils) = nils
  | combines (h1 ::: up t1, h2 ::: up t2) = (h1,h2) ::: up (combines (t1,t2))
  | combines _ = raise Combines;

fun generates f x = x ::: up (generates f (f x));

end;





(*
(* some examples to play with: *)

fun inc n = n + 1;

val nats = (* using a recursive function "from" *)
	let fun from n = n ::: up (from (n+1)) in from 0 end;

val nats2 = (* a direct recursive definition *)
	let val rec z = up (0 ::: up (maps inc (down z))) in down z end;

val nats3 = (* using generates function *)
	generates inc 0;

val evens = maps (fn x => 2*x) nats
and odds  = maps (fn x => 2*x+1) nats;


val fibs = (* using a recursive function "from" *)
	let fun from (x,y) = let val z = x + y in z ::: up (from (y,z)) end
	in 1 ::: up (1 ::: up (from (1,1)))
	end;

val fibs2 = (* direct recursive definition (see e.g. Henderson's book on FP) *)
	let fun adds z1 z2 = maps (op +) (combines (z1,z2))
	    val rec z = up (1 ::: up (1 ::: up (adds (down z) (ts (down z)))))
            in down z  
	(* equivalent alternative:
	    val rec z = up (adds (1 ::: up (1 ::: z)) (1 ::: z))
            in 1 ::: up (1 ::: z) *)
	(* as henderson: 
	    val rec z1 = up (1 ::: z2)
                and z2 = up (1 ::: z3)
		and z3 = up (adds (down z1) (down z2))
	    in down z1 *)
	end;

val fibs3 = (* with generates, reference r remembers the previous integer *)
	let val r = ref 1
	    fun nxt y = let val x = !r in (r := y; x+y) end
	in 1 ::: up (generates nxt 1)
	end;
*)



(* suggested exercises:
consider other possible definitions of lazy lists, e.g.
	datatype 'a lizt = nils | ::: of ('a * 'a lizt) lift;
	exception Hs;
	fun hs nils = raise Hs | hs (::: (up l)) = fst l;
	etc,
*)
