(* mprat.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

multiple precision rational numbers

*)




(* signature 

type mprat
con // : int * int -> mprat
val over : int * int -> mprat
val ~ : mprat -> mprat
val + : mprat * mprat -> mprat
val - : mprat * mprat -> mprat
val * : mprat * mprat -> mprat
val / : mprat -> mprat -> mprat
exception Inv;
val inv : mprat -> mprat
val eq : mprat * mprat -> bool
val > : mprat * mprat -> bool
val < : mprat * mprat -> bool
val >= : mprat * mprat -> bool
val <= : mprat * mprat -> bool
val real'r : mprat -> real
val continued : real -> int -> int list
val mprat'c : int list -> mprat
val mprat'r : real -> mprat
val mprat2'c : int list -> mprat
val mprat2'r : real -> int -> mprat
val real'c : real -> real
val mprat'i : int -> mprat

val string_of_mprat : mprat -> string
exception Mprat_of_string : exn
val mprat_of_string : string -> mprat
val print : mprat -> mprat
val makestring : mprat -> string
*)



(* implementation *)

needs "alist"; (* chop_find *)


structure Mprat =
struct

structure Rep = struct
datatype mprat = // of mp * mp;
end;
infix 8 //;
open Rep;

infix 8 over;
infix 7 * /;
infix 6 + -;
overload ~ + - * / > < >= <=;

local
fun gcd n 00 = n | gcd n m = gcd m (n mod m);
fun zero (a//_) = a = 00;
fun pos (a//_) = a > 00;
in
fun n over d = let val g = if d>=00 then abs(gcd n d) else ~ (abs(gcd n d))
	       in (n div g)//(d div g)
	       end;
fun ~ (n//d) = (~n)//d;
fun (n1//d1) + (n2//d2) = (n1*d2 + d1*n2) over (d1*d2);
fun (x:mprat) - y = x + (~ y);
exception Inv;
fun inv (_//00) = raise Inv | inv (n//d) = d//n;
fun (n1//d1) * (n2//d2) = (n1*n2) over (d1*d2);
fun x / y = x * (inv y);
infix 4 eq;
fun x eq y = zero (x - y);
fun x >  (y:mprat) = pos (x-y);
fun x <  (y:mprat) = y > x;
fun x >= (y:mprat) = not (x < y);
fun x <= (y:mprat) = not (x > y);
end;
nonoverload ~ + - * / > < >= <=;

(* mprat<->real conversions will be provided when mpreal, mpfloor available *)

fun string_of_mprat (x//y) = string_of_mp x^"//"^string_of_mp y;
exception Mprat_of_string;
fun mprat_of_string s =
	let val (l1,l2) = Alist.chop_find (fn x => x ="/") (explode s)
		handle Alist.Chop_find => raise Mprat_of_string
	in case l2 of
	  ("/"::"/"::l) => (mp_of_string (implode l1) // mp_of_string (implode l)
		handle Mp_of_string => raise Mprat_of_string)
	| _ => raise Mprat_of_string
	end;

val makestring = string_of_mprat;
fun print r = (output (std_output(), string_of_mprat r); r);


end; (* structure *)



(* usage:
overload ~ + - * / > < >= <= print makestring;
open Mprat;
nonoverload ~ + - * / > < >= <= print makestring;
infix 8 // over;
infix 4 eq;
*)
