(* slist.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

sets as lists.

*)




(* signature

structure Slist =
   val distinct = - : ''a list -> bool
   val make_set = - : ''a list -> ''a list
   val intersect = - : ''a list -> ''a list -> ''a list
   val subtract = - : ''a list -> ''a list -> ''a list
   val symdiff = - : ''a list -> ''a list -> ''a list
   val union = - : ''a list -> ''a list -> ''a list
   exception Repl : exn
   val repl = - : 'a -> int -> 'a list
   val partition = - : ('a -> bool) -> 'a list -> 'a list * 'a list
   val range = - : int -> int -> int list
   val count = - : ('a -> bool) -> 'a list -> int

*)




structure Slist = (* sets as lists *)
struct 

(* true if all items in list are distinct *)
fun distinct [] = true
  | distinct (h::t) = not (mem h t) andalso distinct t;

(* Makes a list into a set, stripping out duplicate elements *)
fun make_set l = fold (fn a => fn s => if mem a s then s else a::s) l [];

fun intersect l1 l2 = filter (fn x=> mem x l2) l1 
and subtract l1 l2 = filter (fn x=> not (mem x l2)) l1;
fun symdiff l1 l2 = subtract l1 l2 @ subtract l2 l1;
fun union l1 l2 =
  let fun runion [] = l2
        | runion (h::t) = if mem h l2 then runion t else h :: runion t
  in runion l1
  end;
fun union l1 l2 = l1 @ (subtract l2 l1);

(*make the list [x; x; ...; x] of length n*)
exception Repl;
fun repl x n =
	let fun reprec 0 = [] | reprec n = x:: reprec(n-1)
	in if n<0 then raise Repl else reprec n
	end;

(*make the list [from; from+1; ...; to]*)
fun range from to =
    if from>to then []
    else from :: (range (from+1) to);

(* = (l1,l2) in which l1 (l2) holds the items of l (not) satisfying p *)
fun partition p l =
    fold (fn a => fn (y,n) =>
		if p a then ((a::y),n) else (y,(a::n))) l ([],[]);

(* number of elements of l that satisfy p. eqv. length (filter p l) *)
fun count p l =
	let fun cnt [] = 0
	      | cnt (h::t) = if p h then 1 + count p t else count p t
	in cnt l
	end;

end; (* structure *)



