(* stack.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

stacks (functional).

*)




(* interface

structure Stack =

   abstype 'a stack
   exception Pop : exn
   val new   : 'a stack
   val isnew : 'a stack -> bool
   val push  : 'a -> 'a stack -> 'a stack
   val pop   : 'a stack -> 'a * 'a stack
   val top   : 'a stack -> 'a

   abstype 'a istack
   val newstack   : unit -> '1a istack
   val isnewstack : 'a istack -> bool
   val clearstack : 'a istack -> unit
   val pushstack  : 'a -> 'a istack -> unit
   val popstack   : 'a istack -> unit
   val topstack   : 'a istack -> 'a

   type 'a ostack = (unit -> unit) *
		    (unit -> bool) *
		    ('a -> unit) *
		    (unit -> 'a) *
		    (unit -> 'a)
   val newostack : unit -> '1a ostack
 
*)



(* implementation *)

structure Stack =
struct

abstype 'a stack = S of 'a list
with
val new = S [];
fun isnew (S []) = true | isnew _ = false;
fun push x (S s) = S (x::s);
exception Pop;
fun pop (S (h::t)) = (h,S t) | pop _ = raise Pop;
fun top (S (h::t)) = h | top _ = raise Pop;
end;

abstype 'a istack = I of 'a stack ref		(* stacks with states *)
with 
fun newstack () = I (ref new);
fun clearstack (I r) = r := new;
fun isnewstack (I r) = isnew (!r);
fun pushstack x (I r) = r := push x (!r);
fun popstack (I r) = let val (x,s) = pop (!r) in r := s; x end;
fun topstack (I r) = top (!r);
end;

type 'a ostack = 				(* object style stacks *)
		(unit -> unit) *		(* clear method	*)
		(unit -> bool) *		(* isnew 	*)
		('a -> unit) *			(* push		*)
		(unit -> 'a) *			(* pop		*)
		(unit -> 'a);			(* top		*)
fun newostack () : 'a ostack =
	let val r = ref new
	in (fn () => r := new,
	    fn () => isnew (!r),
	    fn x => r := push x (!r),
	    fn () => let val (x,s) = pop (!r) in r := s; x end,
	    fn () => top (!r))
	end;

end; (* structure *)






(* e.g.

open Stack;

(* functional version: *)
val (x,s) = pop (push 456 (push 12 new));
isnew s;

(* or imperative version: *)
val s = newstack () : int istack;
pushstack s 12;
pushstack s 456;
val x = popstack s;
isnewstack s;

*)

