(* tlist.sml
 _______________________________________________________________________
|									|
|			      Projet LCS				|
|			  Groupe OLC, CNRS/LAAS				|
|			 7, Av. du Colonel Roche			|
|			   31077 Toulouse Cedex				|
|_______________________________________________________________________|

tables as lists.

*)




(* signature

structure Tlist =
   exception Range : exn
   val chop = - : int -> 'a list -> 'a list * 'a list
   val nth = - : 'a list -> int -> 'a
   val dropnth = - : int -> 'a list -> 'a list
   val replacenth = - : int -> 'a -> 'a list -> 'a list
   val pack = - : int -> 'a list -> 'a list list
   val transpose = - : 'a list list -> 'a list list

*)


structure Tlist = (* tables as lists (vectors, matrices, etc) *)
struct

exception Range;

(* chops l as two lists (l1,l2), wity length l1 = n *)
fun chop 0 l = ([],l)
  | chop n [] = raise Range
  | chop n (h::t) = let val (l1,l2) = chop (n-1) t in (h::l1,l2) end;

local
fun nthrec [] _ = raise Range
  | nthrec (h::t) n = if n=1 then h else nthrec t (n-1)
in
fun nth l n = if n<0 then raise Range else nthrec l n
end;

fun replacenth n v e =
	let val (l1,_::l2) = chop (n-1) e in l1 @ v::l2 end
	handle Chop => raise Range;

fun dropnth n e =
	let val (l1,_::l2) = chop (n-1) e in l1 @ l2 end
	handle Chop => raise Range;

(* packs list l into a list of rows of n elements *)
fun pack n [] = []
  | pack n l = let val (l1,l2) = chop n l handle Chop => raise Range
	       in l1::pack n l2 end;

(* list of rows => list of colomns *)
fun transpose [] = []     
  | transpose ([]::_) = []
  | transpose m =
        let val (c,m') = split (map (fn (h::l) => (h,l)) m)
        in c::transpose m'
        end;

end; (* structure *)


